///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//  Copyright  NetworkDLS 2002, All rights reserved
//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF 
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A 
// PARTICULAR PURPOSE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef _CMEMPOOL_CPP
#define _CMEMPOOL_CPP
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <Windows.H>
#include <Stdio.H>
#include <Stdlib.H>

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 3
#include <ConIO.H>
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "CMemPool.H"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
#define CMemPoolAssert(sExpression, sMsg) \
	if(sExpression)\
	{ \
		char sAssert[2048]; \
		sprintf_s(sAssert, sizeof(sAssert), \
			"%s\r\n\r\n" \
			"Procedure: %s\r\n" \
			"Source File: %s\r\n" \
			"Build Date/Time: %s (%s)\r\n" \
			"Expression: \"%s\"\r\n" \
			"Line: %d\r\n\r\n", \
			sMsg, __FUNCTION__, __FILE__, __DATE__, __TIME__, #sExpression, __LINE__); \
		this->DbgWrite(sAssert); \
		MessageBox(NULL, sAssert, "CMemPool Exception", MB_ICONSTOP | MB_SYSTEMMODAL); \
	}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CMemPool::~CMemPool(void)
{
	this->Destroy();
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CMemPool::CMemPool(void)
{
	this->Initialize();
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
CMemPool::CMemPool(unsigned int iAllocateIncrement)
{
	this->Initialize(iAllocateIncrement);
}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 2
unsigned int CMemPool::ReservedBytes(void)
{
	this->Lock();
	unsigned int iBytes = this->iReservedBytes;
	this->Unlock();
	return iBytes;
}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CMemPool::IsDebug(void)
{
	#if _MEMPOOL_DEBUG_LEVEL >= 1
		return true;
	#else
		return false;
	#endif
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
	int CMemPool::GetMemFreeCount(void)
	{
		return(this->Items.iAllocated - this->Items.iUsedCount);
	}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
	int CMemPool::GetMemAllocationCount(void)
	{
		return this->Items.iUsedCount;
	}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
	int CMemPool::GetSlotAllocationCount(void)
	{
		return this->Items.iAllocated;
	}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
	bool CMemPool::FreeAll(void)
	{
		bool bResult = true;

		//Free all the memory stored in the [Array] structure.
		for(register unsigned long iSlot = 0; iSlot < this->Items.iAllocated; iSlot++)
		{
			if(!this->Items.Array[iSlot].bFree)
			{
				sprintf_s(this->sDebugText, sizeof(this->sDebugText),
					"Memory leak detected while destroying a managed memory pool item at [0x%X]."
					#if _MEMPOOL_DEBUG_LEVEL >= 2
					"\r\n\r\nAllocated in \"%s\""
					"\r\non line %d for %d bytes."
					#endif
						, this->Items.Array[iSlot].iAddress
					#if _MEMPOOL_DEBUG_LEVEL >= 2
						, this->Items.Array[iSlot].sCodeFile,
						this->Items.Array[iSlot].iLineOfCode,
						this->Items.Array[iSlot].iSize
					#endif						
				);
				CMemPoolAssert("FreeAll", this->sDebugText);

				if(IsDebuggerPresent())
				{
					__debugbreak();
				}

				if(!this->Free((void *)this->Items.Array[iSlot].iAddress))
				{
					bResult = false;
				}
			}
		}

		return bResult;
	}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
	int CMemPool::GetSlotByAddress(const unsigned long iAddress)
	{
		for(register unsigned long iSlot = 0; iSlot < this->Items.iAllocated; iSlot++)
		{
			if(!this->Items.Array[iSlot].bFree)
			{
				if(this->Items.Array[iSlot].iAddress == iAddress)
				{
					return iSlot;
				}
			}
		}

		sprintf_s(this->sDebugText, sizeof(this->sDebugText),
			"Application attempted to access an unmanaged memory address at [0x%X].", iAddress);
		CMemPoolAssert("GetSlotByAddress", this->sDebugText);

		if(IsDebuggerPresent())
		{
			__debugbreak();
		}

		return -1; //Did not find the address specified.
	}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
	int CMemPool::FindFreeAllocSlot(void)
	{
		for(register unsigned long iSlot = 0; iSlot < this->Items.iAllocated; iSlot++)
		{
			if(this->Items.Array[iSlot].bFree)
			{
				return iSlot;
			}
		}

		unsigned long iMemory = (sizeof(MEMITEM) * this->Items.iAllocated)
			+ (sizeof(MEMITEM) * this->Items.iAllocIncrement);

		if((this->Items.Array = (MEMITEM *) realloc(this->Items.Array, iMemory)) == NULL)
		{
			sprintf_s(this->sDebugText, sizeof(this->sDebugText),
				"Failed to reallocate the managed Array structure for [%d] bytes.", iMemory);
			CMemPoolAssert("FindFreeAllocSlot", this->sDebugText);

			if(IsDebuggerPresent())
			{
				__debugbreak();
			}

			return -1;
		}

		this->Items.iAllocated = (this->Items.iAllocated + this->Items.iAllocIncrement);

		return this->FindFreeAllocSlot();
	}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
	bool CMemPool::Initialize(unsigned int iAllocateIncrement)
	{
		InitializeCriticalSection(&this->csMemLock);

		memset(&this->Items, 0, sizeof(Items));
		this->Items.iAllocIncrement = iAllocateIncrement;

		#if _MEMPOOL_DEBUG_LEVEL > 2
			this->iReservedBytes = 0;

			#if _MEMPOOL_DEBUG_LEVEL >= 3
				this->hStdOut = NULL;
			#endif
		#endif

		if((this->Items.Array = (MEMITEM *)
			calloc(sizeof(MEMITEM), this->Items.iAllocIncrement)) == NULL)
		{
			CMemPoolAssert("Initialize", "Memory pool Failed to initialize.");

			if(IsDebuggerPresent())
			{
				__debugbreak();
			}

			DeleteCriticalSection(&this->csMemLock);
			return false;
		}

		this->Items.iAllocated = this->Items.iAllocIncrement;

		for(register unsigned long iSlot = 0; iSlot < this->Items.iAllocated; iSlot++)
		{
			this->Items.Array[iSlot].bFree = true;
			this->Items.Array[iSlot].iAddress = 0;
			#if _MEMPOOL_DEBUG_LEVEL > 2
				this->Items.Array[iSlot].iSize = 0;
				this->Items.Array[iSlot].iLineOfCode = -1;
				this->Items.Array[iSlot].sCodeFile = NULL;
			#endif
		}

		#if _MEMPOOL_DEBUG_LEVEL >= 3
			return this->OpenConsole();
		#else
			return true;
		#endif
	}
#endif
	
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
bool CMemPool::Initialize(void)
{
	#if _MEMPOOL_DEBUG_LEVEL >= 1
		return this->Initialize(100);
	#else
		InitializeCriticalSectionAndSpinCount(&this->csMemLock, 4000);
		return true;
	#endif
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
	bool CMemPool::Destroy(void)
	{
		bool bResult = false;

		this->Lock();

		bResult = this->FreeAll();

		free(this->Items.Array);
			
		this->Unlock();

		DeleteCriticalSection(&this->csMemLock);

		#if _MEMPOOL_DEBUG_LEVEL >= 3
			bResult = bResult && this->CloseConsole();
		#endif

		return bResult;
	}
#else
	void CMemPool::Destroy(void)
	{
		DeleteCriticalSection(&this->csMemLock);
	}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CMemPool::Lock(void)
{
	EnterCriticalSection(&this->csMemLock);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CMemPool::Unlock(void)
{
	LeaveCriticalSection(&this->csMemLock);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
	bool CMemPool::Free(void *lpMemory)
	{
		this->Lock();

		#if _MEMPOOL_DEBUG_LEVEL >= 3
			sprintf_s(this->sDebugText, sizeof(this->sDebugText), "Freeing address 0x%X\n", lpMemory);
			this->DbgWrite(this->sDebugText);
		#endif

		unsigned long iSlot = 0;

		if(lpMemory == NULL)
		{
			CMemPoolAssert("Free", "Application attempted to free a NULL pointer.");

			if(IsDebuggerPresent())
			{
				__debugbreak();
			}

			this->Unlock();
			return false;
		}

		if((iSlot = this->GetSlotByAddress((int)lpMemory)) >= 0)
		{
			//Free the memory that the user allocated, also mark the [Array] slot as free.
			free(lpMemory);
			
			lpMemory = NULL;

			#if _MEMPOOL_DEBUG_LEVEL >= 2
				free(this->Items.Array[iSlot].sCodeFile);
				this->iReservedBytes -= this->Items.Array[iSlot].iSize;
				this->Items.Array[iSlot].sCodeFile = NULL;
				this->Items.Array[iSlot].iLineOfCode = 0;
				this->Items.Array[iSlot].iSize = 0;
			#endif

			this->Items.Array[iSlot].bFree = true;
			this->Items.Array[iSlot].iAddress = 0;

			this->Items.iUsedCount--;
		}
		else{
			sprintf_s(this->sDebugText, sizeof(this->sDebugText),
				"Address [0x%X] not found in the managed memory pool.", lpMemory);
			CMemPoolAssert("Free", this->sDebugText);

			if(IsDebuggerPresent())
			{
				__debugbreak();
			}

			this->Unlock();
			return false;
		}

		this->Unlock();
		return true;
	}
#else
	void CMemPool::Free(void *lpMemory)
	{
		this->Lock();
		free(lpMemory);
		this->Unlock();
	}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
	#if _MEMPOOL_DEBUG_LEVEL >= 2
		void *CMemPool::_dbg_ReAllocate(void *lpMemory, const unsigned long iSize, const unsigned long iCount,
						const char *sCodeFile, const int iLineOfCode)
	#else
		void *CMemPool::ReAllocate(void *lpMemory, const unsigned long iSize, const unsigned long iCount)
	#endif
	{
		if(lpMemory)
		{
			this->Lock();

			unsigned long iSlot = 0;
			char *sMemory = NULL;

			if((iSlot = this->GetSlotByAddress((int)lpMemory)) >= 0)
			{
				if((sMemory = (char *) realloc(lpMemory, iSize * iCount)) == NULL)
				{
					sprintf_s(this->sDebugText, sizeof(this->sDebugText),
						"Failed to reallocate managed user memory for [%d] bytes.", iSize * iCount);
					CMemPoolAssert("ReAllocate", this->sDebugText);

					if(IsDebuggerPresent())
					{
						__debugbreak();
					}

					this->Unlock();
					return NULL;
				}

				#if _MEMPOOL_DEBUG_LEVEL >= 3
					sprintf_s(this->sDebugText, sizeof(this->sDebugText),
						"Realloc %d bytes. (0x%X -> 0x%X) @ %s:%d\n",
						iSize * iCount, lpMemory, sMemory,
						this->Items.Array[iSlot].sCodeFile,
						this->Items.Array[iSlot].iLineOfCode);
					this->DbgWrite(this->sDebugText);
				#endif

				this->Items.Array[iSlot].bFree = false;
				this->Items.Array[iSlot].iAddress = (int) sMemory;
				#if _MEMPOOL_DEBUG_LEVEL >= 2
					this->iReservedBytes -= this->Items.Array[iSlot].iSize;
					this->Items.Array[iSlot].iSize = iSize * iCount;
					this->iReservedBytes += this->Items.Array[iSlot].iSize;
				#endif
			}
			else{
				sprintf_s(this->sDebugText, sizeof(this->sDebugText),
					"Address [0x%X] not found in the managed memory pool.", lpMemory);
				CMemPoolAssert("ReAllocate", this->sDebugText);

				if(IsDebuggerPresent())
				{
					__debugbreak();
				}

				this->Unlock();
				return NULL;
			}

			this->Unlock();
			return sMemory;	
		}
		else{
			#if _MEMPOOL_DEBUG_LEVEL >= 2
				return this->_dbg_Allocate(iSize, iCount, sCodeFile, iLineOfCode);
			#else
				return this->Allocate(iSize, iCount);
			#endif
		}
	}
#else
	void *CMemPool::ReAllocate(void *lpMemory, const unsigned long iSize, const unsigned long iCount)
	{
		this->Lock();
		char *sMemory = (char *) realloc(lpMemory, iSize * iCount);
		this->Unlock();

		return sMemory;
	}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
	#if _MEMPOOL_DEBUG_LEVEL >= 2
		void *CMemPool::_dbg_Allocate(const unsigned long iSize, const unsigned long iCount,
		const char *sCodeFile, const int iLineOfCode)
	#else
		void *CMemPool::Allocate(const unsigned long iSize, const unsigned long iCount)
	#endif
	{
		this->Lock();

		char *sMemory = NULL;
		unsigned long iSlot = 0;

		if((sMemory = (char *) calloc(iSize, iCount)) == NULL)
		{
			sprintf_s(this->sDebugText, sizeof(this->sDebugText),
				"Failed to allocate managed user memory for [%d] bytes.", iSize * iCount);
			CMemPoolAssert("Allocate", this->sDebugText);

			if(IsDebuggerPresent())
			{
				__debugbreak();
			}

			this->Unlock();
			return NULL;
		}

		if((iSlot = this->FindFreeAllocSlot()) < 0)
		{
			sprintf_s(this->sDebugText, sizeof(this->sDebugText),
				"Failed to find qualified memory slot for [%d] bytes.", iSize * iCount);
			CMemPoolAssert("Allocate", this->sDebugText);

			if(IsDebuggerPresent())
			{
				__debugbreak();
			}

			this->Unlock();
			return NULL;
		}

		this->Items.Array[iSlot].iAddress = (int) sMemory;
		this->Items.Array[iSlot].bFree = false;

		#if _MEMPOOL_DEBUG_LEVEL >= 2
			this->Items.Array[iSlot].iSize = (iCount * iSize);
			this->Items.Array[iSlot].iLineOfCode = iLineOfCode;
			this->iReservedBytes += this->Items.Array[iSlot].iSize;

			int iCodeFileAlloc = this->GetFileName(sCodeFile, sTempFile) + 1;
			this->Items.Array[iSlot].sCodeFile = (char *) calloc(iCodeFileAlloc, sizeof(char));
			memcpy_s(this->Items.Array[iSlot].sCodeFile, iCodeFileAlloc, sTempFile, iCodeFileAlloc);

			#if _MEMPOOL_DEBUG_LEVEL >= 3
					sprintf_s(this->sDebugText, sizeof(this->sDebugText),
						"Alloc %d bytes. (0x%X) @ %s:%d\n",
						iSize * iCount, sMemory,
						this->Items.Array[iSlot].sCodeFile,
						this->Items.Array[iSlot].iLineOfCode);
					this->DbgWrite(this->sDebugText);
			#endif
		#endif

		this->Items.iUsedCount++;

		this->Unlock();
		return sMemory;
	}
#else
	void *CMemPool::Allocate(const unsigned long iSize, const unsigned long iCount)
	{
		this->Lock();
		char *sMemory = (char *) calloc(iCount, iSize);
		this->Unlock();

		return sMemory;
	}
#endif

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 3
BOOL WINAPI CMemPool_ConsoleHandler(DWORD dwCtrlType)
{
    if(dwCtrlType == CTRL_CLOSE_EVENT)
    {
		return true;
    }

    return FALSE;
}
#endif

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 3
bool CMemPool::OpenConsole(void)
{
    if(this->hStdOut == NULL)
	{
		if(AllocConsole())
		{
			SetConsoleCtrlHandler(CMemPool_ConsoleHandler, TRUE);
			SetConsoleTitle("CMemPool");
		}

		if((this->hStdOut = GetStdHandle(STD_OUTPUT_HANDLE)) != NULL)
		{
			return true;
		}
	}
    return false;
}
#endif

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 3
bool CMemPool::CloseConsole(void)
{
	if(this->hStdOut)
	{
		if(FreeConsole())
	    {
	        CloseHandle(this->hStdOut);
			this->hStdOut = NULL;
	        return true;
		}
    }
    return false;
}
#endif

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
int CMemPool::GetFileName(const char *sInPath, char *sOutFile)
{
	int iWPos = 0;

	for(int iPos = (int)strlen(sInPath) - 1;
		iPos >= 0 && sInPath[iPos] != '\\' && sInPath[iPos] != '/'; iPos--)
	{
		sOutFile[iWPos++] = sInPath[iPos];
	}

	sOutFile[iWPos] = '\0';

	char *String1 = NULL;
	char *String2 = NULL;

	for(String1 = sOutFile, String2 = sOutFile + iWPos - 1;
		String2 > String1;
		++String1, --String2)
	{
		*String1 ^= *String2;
		*String2 ^= *String1;
		*String1 ^= *String2;
	}

	return iWPos;
}
#endif

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 3
void CMemPool::DbgWrite(const int iValue)
{
	char sVal[64];
	_itoa_s(iValue, sVal, sizeof(sVal), 10);
	this->DbgWrite(sVal);
}
#endif

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _MEMPOOL_DEBUG_LEVEL >= 1
void CMemPool::DbgWrite(const char *sBuffer)
{
	#if _MEMPOOL_DEBUG_LEVEL >= 3
	if(this->hStdOut)
	{
		DWORD dwWritten = 0;
		WriteFile(this->hStdOut, sBuffer, (int)strlen(sBuffer), &dwWritten, 0);
    }
	#endif
}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif
